r"""Tests for the lampe.plots module."""

import pytest
import numpy as np

from lampe.plots import *


def test_nice_rc():
    assert type(nice_rc()) is dict


def test_corner():
    # Samples
    samples = np.random.randn(2**16, 3)
    labels = ['a', 'b', 'c']
    figure = corner(samples, bins=32, labels=labels, figsize=(4.8, 4.8))

    assert np.asarray(figure.axes).size == 3**2

    # Histograms
    hists = [[None] * 3 for _ in range(3)]

    for i in range(3):
        hists[i][i] = np.random.rand(8 + 2 * i)

        for j in range(i):
            hists[i][j] = np.random.rand(8 + 2 * i, 8 + 2 * j)

    new = corner(hists, (-np.ones(3), np.ones(3)), figure=figure)

    assert figure is new

    # Mark points
    mark_point(figure, [0.5, 0.3, -0.7], color='black')


def test_coverage_plot():
    levels = np.random.rand(512) ** 2
    coverages = np.linspace(0, 1, 512)
    figure = coverage_plot(levels, coverages)



r"""Tests for the lampe.utils module."""

import math
import pytest
import torch

from lampe.utils import *
from torch import rand


def test_GDStep():
    layer = torch.nn.Linear(3, 3)
    optimizer = torch.optim.SGD(layer.parameters(), lr=1e-3)
    step = GDStep(optimizer)

    # Normal step
    before = layer.weight.detach().clone()

    x = rand(3)
    y = layer(x)

    loss = (y - x).square().sum()
    loss = step(loss)

    after = layer.weight.detach().clone()

    assert not loss.requires_grad
    assert not torch.allclose(before, after)

    # Detached loss
    with pytest.raises(RuntimeError):
        step(loss)

    # Non-scalar loss
    with pytest.raises(RuntimeError):
        step(layer(x))

    # Non-finite loss
    before = layer.weight.detach().clone()

    x = rand(3) * float('inf')
    y = layer(x)

    loss = (y - x).square().sum()
    loss = step(loss)

    after = layer.weight.detach().clone()

    assert torch.allclose(before, after)

    # Non-finite grad with clip
    step.clip = 1

    before = layer.weight.detach().clone()

    x = rand(3)
    y = layer(x)

    loss = (y.sum() * 0).sqrt()
    loss = step(loss)

    after = layer.weight.detach().clone()

    assert torch.allclose(before, after)

    # Non-finite grad without clip
    step.clip = None

    x = rand(3)
    y = layer(x)

    loss = (y.sum() * 0).sqrt()
    loss = step(loss)

    assert torch.isnan(layer.weight).all()


def test_gridapply():
    f = lambda x: x.square().sum(dim=-1)
    lower, upper = torch.zeros(3), torch.ones(3)

    x, y = gridapply(f, (lower, upper), bins=8)

    assert x.shape == (8, 8, 8, 3)
    assert (lower <= x).all() and (x <= upper).all()
    assert y.shape == (8, 8, 8)
    assert torch.allclose(f(x), y)



r"""Tests for the lampe.data module."""

import h5py
import numpy as np
import pytest
import torch

from lampe.data import *


def test_JointLoader():
    prior = torch.distributions.Normal(torch.zeros(3), torch.ones(3))

    count = 0

    def simulator(theta):
        nonlocal count

        count += 1

        assert torch.is_tensor(theta)
        assert theta.dtype == torch.float

        return torch.repeat_interleave(theta, 2, dim=-1)

    # Non-vectorized
    loader = JointLoader(prior, simulator, batch_size=4, vectorized=False)

    assert isinstance(loader, torch.utils.data.DataLoader)

    it = iter(loader)
    theta1, x1 = next(it)
    theta2, x2 = next(it)

    assert count == 8

    assert torch.is_tensor(theta1) and torch.is_tensor(x1)
    assert theta1.shape == (4, 3) and x1.shape == (4, 6)
    assert theta1.dtype == x1.dtype == torch.float
    assert (theta1 != theta2).all()

    # Vectorized
    loader = JointLoader(prior, simulator, batch_size=5, vectorized=True)

    it = iter(loader)
    theta1, x1 = next(it)
    theta2, x2 = next(it)

    assert count == 10

    assert torch.is_tensor(theta1) and torch.is_tensor(x1)
    assert theta1.shape == (5, 3) and x1.shape == (5, 6)
    assert theta1.dtype == x1.dtype == torch.float
    assert (theta1 != theta2).all()

    # NumPy
    def simulator(theta):
        assert not torch.is_tensor(theta)
        assert theta.dtype == np.double

        return np.repeat(theta, 3, axis=-1)

    loader = JointLoader(prior, simulator, batch_size=6, numpy=True)

    it = iter(loader)
    theta1, x1 = next(it)

    assert torch.is_tensor(theta1) and torch.is_tensor(x1)
    assert theta1.shape == (6, 3) and x1.shape == (6, 9)
    assert theta1.dtype == x1.dtype == torch.float


def test_JointDataset():
    theta, x = torch.randn(1024, 5), torch.randn(1024, 16, 2)

    # Index
    dataset = JointDataset(theta, x)

    for i in (0, 1, 2, 8, 32, 128, -1):
        theta_i, x_i = dataset[i]

        assert (theta_i == theta[i]).all()
        assert (x_i == x[i]).all()

    # Iter
    for i, (theta_i, x_i) in enumerate(dataset):
        assert (theta_i == theta[i]).all()
        assert (x_i == x[i]).all()

    assert i == len(dataset) - 1 == len(theta) - 1

    # Shuffle
    dataset = JointDataset(theta, x, batch_size=256, shuffle=True)

    it = iter(dataset)
    theta1, x1 = next(it)
    theta2, x2 = next(it)

    assert len(theta1) == len(theta2) == 256
    assert len(x1) == len(x2) == 256

    theta3, x3 = torch.cat((theta1, theta2)), torch.cat((x1, x2))

    match = (theta3[:, None, :] == theta).all(dim=-1)

    assert (match.sum(dim=0) <= 1).all()
    assert (match.sum(dim=-1) == 1).all()

    _, index = torch.nonzero(match, as_tuple=True)

    assert (x3 == x[index]).all()


def test_H5Dataset(tmp_path):
    prior = torch.distributions.Normal(torch.zeros(3), torch.ones(3))
    simulator = lambda theta: torch.repeat_interleave(theta, 2, dim=-1)

    theta = prior.sample((4096,))
    x = simulator(theta)

    # Store
    pairs = list(zip(theta.split(256), x.split(256)))

    H5Dataset.store(pairs, tmp_path / 'data_1.h5', size=4096)
    H5Dataset.store(iter(pairs), tmp_path / 'data_2.h5', size=4096)
    H5Dataset.store([(theta, x)], tmp_path / 'data_3.h5', size=256)

    with pytest.raises(FileExistsError):
        H5Dataset.store(pairs, tmp_path / 'data_1.h5', size=4096)

    H5Dataset.store(pairs, tmp_path / 'data_1.h5', overwrite=True, size=4096)

    # Load
    for file in tmp_path.glob('data_*.h5'):
        dataset = H5Dataset(file)

        assert len(dataset) in {256, 4096}

        ## __getitem__
        for i in (0, 1, 2, 8, 32, 128):
            theta_i, x_i = dataset[i]

            assert (theta_i == theta[i]).all()
            assert (x_i == x[i]).all()

        ## __iter__
        for i, (theta_i, x_i) in enumerate(dataset):
            assert (theta_i == theta[i]).all()
            assert (x_i == x[i]).all()

        assert i == len(dataset) - 1

    # Shuffle
    dataset = H5Dataset(tmp_path / 'data_1.h5', batch_size=256, shuffle=True)

    it = iter(dataset)
    theta1, x1 = next(it)
    theta2, x2 = next(it)

    assert len(theta1) == len(theta2) == 256
    assert len(x1) == len(x2) == 256

    theta3, x3 = torch.cat((theta1, theta2)), torch.cat((x1, x2))

    match = (theta3[:, None, :] == theta).all(dim=-1)

    assert (match.sum(dim=0) <= 1).all()
    assert (match.sum(dim=-1) == 1).all()

    _, index = torch.nonzero(match, as_tuple=True)

    assert (x3 == x[index]).all()

    # Load in memory
    new = dataset.to_memory()

    assert len(dataset) == len(new)
    assert isinstance(new, JointDataset)



r"""Inference components such as estimators, training losses and MCMC samplers."""

__all__ = [
    'NRE',
    'NRELoss',
    'BNRELoss',
    'CNRELoss',
    'BCNRELoss',
    'AMNRE',
    'AMNRELoss',
    'NPE',
    'NPELoss',
    'FMPE',
    'FMPELoss',
    'MetropolisHastings',
]

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from itertools import islice
from torch import Tensor, BoolTensor, Size
from typing import *

from zuko.distributions import Distribution, DiagNormal, NormalizingFlow
from zuko.flows import FlowModule, MAF
from zuko.transforms import FreeFormJacobianTransform
from zuko.utils import broadcast

from .nn import MLP


class NRE(nn.Module):
    r"""Creates a neural ratio estimation (NRE) network.

    The principle of neural ratio estimation is to train a classifier network
    :math:`d_\phi(\theta, x)` to discriminate between pairs :math:`(\theta, x)` equally
    sampled from the joint distribution :math:`p(\theta, x)` and the product of the
    marginals :math:`p(\theta)p(x)`. Formally, the optimization problem is

    .. math:: \arg\min_\phi
        \frac{1}{2} \mathbb{E}_{p(\theta, x)} \big[ \ell(d_\phi(\theta, x)) \big] +
        \frac{1}{2} \mathbb{E}_{p(\theta)p(x)} \big[ \ell(1 - d_\phi(\theta, x)) \big]

    where :math:`\ell(p) = -\log p` is the negative log-likelihood. For this task, the
    decision function modeling the Bayes optimal classifier is

    .. math:: d(\theta, x) = \frac{p(\theta, x)}{p(\theta, x) + p(\theta) p(x)}

    thereby defining the likelihood-to-evidence (LTE) ratio

    .. math:: r(\theta, x)
        = \frac{d(\theta, x)}{1 - d(\theta, x)}
        = \frac{p(\theta, x)}{p(\theta) p(x)}
        = \frac{p(x | \theta)}{p(x)}
        = \frac{p(\theta | x)}{p(\theta)} .

    To prevent numerical stability issues when :math:`d_\phi(\theta, x) \to 0`,
    the neural network returns the logit of the class prediction
    :math:`\text{logit}(d_\phi(\theta, x)) = \log r_\phi(\theta, x)`.

    References:
        | Approximating Likelihood Ratios with Calibrated Discriminative Classifiers (Cranmer et al., 2015)
        | https://arxiv.org/abs/1506.02169

        | Likelihood-free MCMC with Amortized Approximate Ratio Estimators (Hermans et al., 2019)
        | https://arxiv.org/abs/1903.04057

    Arguments:
        theta_dim: The dimensionality :math:`D` of the parameter space.
        x_dim: The dimensionality :math:`L` of the observation space.
        build: The network constructor (e.g. :class:`lampe.nn.ResMLP`). It takes the
            number of input and output features as positional arguments.
        kwargs: Keyword arguments passed to the constructor.
    """

    def __init__(
        self,
        theta_dim: int,
        x_dim: int,
        build: Callable[[int, int], nn.Module] = MLP,
        **kwargs,
    ):
        super().__init__()

        self.net = build(theta_dim + x_dim, 1, **kwargs)

    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(*, D)`.
            x: The observation :math:`x`, with shape :math:`(*, L)`.

        Returns:
            The log-ratio :math:`\log r_\phi(\theta, x)`, with shape :math:`(*,)`.
        """

        theta, x = broadcast(theta, x, ignore=1)

        return self.net(torch.cat((theta, x), dim=-1)).squeeze(-1)


class NRELoss(nn.Module):
    r"""Creates a module that calculates the cross-entropy loss for a NRE network.

    Given a batch of :math:`N \geq 2` pairs :math:`(\theta_i, x_i)`, the module returns

    .. math:: l = \frac{1}{2N} \sum_{i = 1}^N
        \ell(d_\phi(\theta_i, x_i)) + \ell(1 - d_\phi(\theta_{i+1}, x_i))

    where :math:`\ell(p) = -\log p` is the negative log-likelihood.

    Arguments:
        estimator: A log-ratio network :math:`\log r_\phi(\theta, x)`.
    """

    def __init__(self, estimator: nn.Module):
        super().__init__()

        self.estimator = estimator

    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(N, D)`.
            x: The observation :math:`x`, with shape :math:`(N, L)`.

        Returns:
            The scalar loss :math:`l`.
        """

        theta_prime = torch.roll(theta, 1, dims=0)

        log_r, log_r_prime = self.estimator(
            torch.stack((theta, theta_prime)),
            x,
        )

        l1 = -F.logsigmoid(log_r).mean()
        l0 = -F.logsigmoid(-log_r_prime).mean()

        return (l1 + l0) / 2


class BNRELoss(nn.Module):
    r"""Creates a module that calculates the balanced cross-entropy loss for a balanced
    NRE (BNRE) network.

    Given a batch of :math:`N \geq 2` pairs :math:`(\theta_i, x_i)`, the module returns

    .. math::
        l & = \frac{1}{2N} \sum_{i = 1}^N
            \ell(d_\phi(\theta_i, x_i)) + \ell(1 - d_\phi(\theta_{i+1}, x_i)) \\
          & + \lambda \left(1 - \frac{1}{N} \sum_{i = 1}^N
            d_\phi(\theta_i, x_i) + d_\phi(\theta_{i+1}, x_i) \right)^2

    where :math:`\ell(p) = -\log p` is the negative log-likelihood.

    References:
        | Towards Reliable Simulation-Based Inference with Balanced Neural Ratio Estimation (Delaunoy et al., 2022)
        | https://arxiv.org/abs/2208.13624

    Arguments:
        estimator: A log-ratio network :math:`\log r_\phi(\theta, x)`.
        lmbda: The weight :math:`\lambda` controlling the strength of the balancing
            condition.
    """

    def __init__(self, estimator: nn.Module, lmbda: float = 100.0):
        super().__init__()

        self.estimator = estimator
        self.lmbda = lmbda

    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(N, D)`.
            x: The observation :math:`x`, with shape :math:`(N, L)`.

        Returns:
            The scalar loss :math:`l`.
        """

        theta_prime = torch.roll(theta, 1, dims=0)

        log_r, log_r_prime = self.estimator(
            torch.stack((theta, theta_prime)),
            x,
        )

        l1 = -F.logsigmoid(log_r).mean()
        l0 = -F.logsigmoid(-log_r_prime).mean()
        lb = (torch.sigmoid(log_r) + torch.sigmoid(log_r_prime) - 1).mean().square()

        return (l1 + l0) / 2 + self.lmbda * lb


class CNRELoss(nn.Module):
    r"""Creates a module that calculates the cross-entropy loss for a contrastive NRE
    (CNRE) network.

    The principle of contrastive neural ratio estimation (CNRE) is to predict whether a
    set :math:`\Theta = \{\theta^1, \dots, \theta^K\}` contains or not the parameters
    that originated an observation :math:`x`. The elements of :math:`\Theta` are drawn
    independently from the prior :math:`p(\theta)` and the element :math:`\theta^k` that
    originates the observation :math:`x \sim p(x | \theta^k)` is chosen uniformly within
    :math:`\Theta`, such that

    .. math:: p(\Theta, x)
        & = p(\Theta) \, p(x | \Theta) \\
        & = p(\Theta) \frac{1}{K} \sum_{k = 1}^K p(x | \theta^k) \\
        & = p(\Theta) \, p(x) \frac{1}{K} \sum_{k = 1}^K r(\theta^k, x)

    where :math:`r(\theta, x)` is the likelihood-to-evidence (LTE) ratio. The task
    is to discriminate between pairs :math:`(\Theta, x)` for which :math:`\Theta` either
    does or does not contain the nominal parameters of :math:`x`, similar to the
    original NRE optimization problem. For this task, the decision function
    modeling the Bayes optimal classifier is

    .. math:: d(\Theta, x)
        = \frac{p(\Theta, x)}{p(\Theta, x) + \frac{1}{\gamma} p(\Theta) p(x)}
        = \frac{\sum_{k = 1}^K r(\theta^k, x)}{\frac{K}{\gamma} + \sum_{k = 1}^K r(\theta^k, x)} \, ,

    where :math:`\gamma \in \mathbb{R}^+` are the odds of :math:`\Theta` containing to
    not containing the nominal parameters. Consequently, a classifier
    :math:`d_\phi(\Theta, x)` can be equivalently replaced and trained as a composition
    of ratios :math:`r_\phi(\theta^k, x)`. Eventually, given a batch of :math:`N \geq
    2K` pairs :math:`(\theta_i, x_i)`, the module returns

    .. math:: l = \frac{1}{N} \sum_{i = 1}^N
        \frac{\gamma}{\gamma + 1} \ell(d_\phi(\Theta_i, x_i))
        + \frac{1}{\gamma + 1} \ell(1 - d_\phi(\Theta_{i+K}, x_i))

    where :math:`\ell(p) = -\log p` is the negative log-likelihood and :math:`\Theta_i =
    \{\theta_i, \dots, \theta_{i+K-1}\}`.

    Note:
        The quantity :math:`d_\phi(\Theta, x)` corresponds to :math:`q_\phi(y \neq 0 |
        \Theta, x)` or :math:`1 - q_\phi(y = 0 | \Theta, x)` in the notations of Miller
        et al. (2022).

    References:
        | Contrastive Neural Ratio Estimation (Miller et al., 2022)
        | https://arxiv.org/abs/2210.06170

    Arguments:
        estimator: A log-ratio network :math:`\log r_\phi(\theta, x)`.
        cardinality: The cardinality :math:`K` of :math:`\Theta`.
        gamma: The odds ratio :math:`\gamma`.
    """

    def __init__(
        self,
        estimator: nn.Module,
        cardinality: int = 2,
        gamma: float = 1.0,
    ):
        super().__init__()

        self.estimator = estimator
        self.cardinality = cardinality
        self.gamma = gamma

    def logits(self, theta: Tensor, x: Tensor) -> Tensor:
        theta = torch.cat((theta, theta[:self.cardinality - 1]), dim=0)
        theta = theta.unfold(0, self.cardinality, 1)
        theta = theta.movedim(-1, 0)

        theta_prime = torch.roll(theta, self.cardinality, dims=1)

        log_r, log_r_prime = self.estimator(
            torch.stack((theta, theta_prime)),
            x,
        )

        shift = math.log(self.gamma / self.cardinality)

        log_r = torch.logsumexp(log_r, dim=0) + shift
        log_r_prime = torch.logsumexp(log_r_prime, dim=0) + shift

        return log_r, log_r_prime

    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(N, D)`.
            x: The observation :math:`x`, with shape :math:`(N, L)`.

        Returns:
            The scalar loss :math:`l`.
        """

        log_r, log_r_prime = self.logits(theta, x)

        l1 = -F.logsigmoid(log_r).mean()
        l0 = -F.logsigmoid(-log_r_prime).mean()

        return (self.gamma * l1 + l0) / (self.gamma + 1)


class BCNRELoss(CNRELoss):
    r"""Creates a module that calculates the balanced cross-entropy loss for a balanced
    CNRE (BCNRE) network.

    Given a batch of :math:`N \geq 2K` pairs :math:`(\theta_i, x_i)`, the module returns

    .. math::
        l & = \frac{1}{N} \sum_{i = 1}^N
            \frac{\gamma}{\gamma + 1} \ell(d_\phi(\Theta_i, x_i))
            + \frac{1}{\gamma + 1} \ell(1 - d_\phi(\Theta_{i+K}, x_i)) \\
          & + \lambda \left(1 - \frac{1}{N} \sum_{i = 1}^N
            d_\phi(\Theta_i, x_i) + d_\phi(\Theta_{i+K}, x_i) \right)^2

    where :math:`\ell(p) = -\log p` is the negative log-likelihood and :math:`\Theta_i =
    \{\theta_i, \dots, \theta_{i+K-1}\}`.

    References:
        | Balancing Simulation-based Inference for Conservative Posteriors (Delaunoy et al., 2023)
        | https://arxiv.org/abs/2304.10978

    Arguments:
        estimator: A log-ratio network :math:`\log r_\phi(\theta, x)`.
        cardinality: The cardinality :math:`K` of :math:`\Theta`.
        gamma: The odds ratio :math:`\gamma`.
        lmbda: The weight :math:`\lambda` controlling the strength of the balancing
            condition.
    """

    def __init__(
        self,
        estimator: nn.Module,
        cardinality: int = 2,
        gamma: float = 1.0,
        lmbda: float = 100.0,
    ):
        super().__init__(estimator, cardinality, gamma)

        self.lmbda = lmbda

    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(N, D)`.
            x: The observation :math:`x`, with shape :math:`(N, L)`.

        Returns:
            The scalar loss :math:`l`.
        """

        log_r, log_r_prime = self.logits(theta, x)

        l1 = -F.logsigmoid(log_r).mean()
        l0 = -F.logsigmoid(-log_r_prime).mean()
        lb = (torch.sigmoid(log_r) + torch.sigmoid(log_r_prime) - 1).mean().square()

        return (self.gamma * l1 + l0) / (self.gamma + 1) + self.lmbda * lb


class AMNRE(NRE):
    r"""Creates an arbitrary marginal neural ratio estimation (AMNRE) network.

    The principle of AMNRE is to introduce, as input to the classifier, a binary mask
    :math:`b \in \{0, 1\}^D` indicating a subset of parameters :math:`\theta_b =
    (\theta_i: b_i = 1)` of interest. Intuitively, this allows the classifier to
    distinguish subspaces and to learn a different ratio for each of them. Formally, the
    classifier network takes the form :math:`d_\phi(\theta_b, x, b)` and the
    optimization problem becomes

    .. math:: \arg\min_\phi
        \frac{1}{2} \mathbb{E}_{p(\theta, x) P(b)} \big[ \ell(d_\phi(\theta_b, x, b)) \big] +
        \frac{1}{2} \mathbb{E}_{p(\theta)p(x) P(b)} \big[ \ell(1 - d_\phi(\theta_b, x, b)) \big],

    where :math:`P(b)` is a binary mask distribution. In this context, the Bayes
    optimal classifier is

    .. math:: d(\theta_b, x, b)
        = \frac{p(\theta_b, x)}{p(\theta_b, x) + p(\theta_b) p(x)}
        = \frac{r(\theta_b, x)}{1 + r(\theta_b, x)} .

    Therefore, a classifier network trained for AMNRE gives access to an estimator
    :math:`\log r_\phi(\theta_b, x, b)` of all marginal LTE log-ratios
    :math:`\log r(\theta_b, x)`.

    References:
        | Arbitrary Marginal Neural Ratio Estimation for Simulation-based Inference (Rozet et al., 2021)
        | https://arxiv.org/abs/2110.00449

    Arguments:
        theta_dim: The dimensionality :math:`D` of the parameter space.
        x_dim: The dimensionality :math:`L` of the observation space.
        args: Positional arguments passed to :class:`NRE`.
        kwargs: Keyword arguments passed to :class:`NRE`.
    """

    def __init__(
        self,
        theta_dim: int,
        x_dim: int,
        *args,
        **kwargs,
    ):
        super().__init__(theta_dim, x_dim + theta_dim, *args, **kwargs)

    def forward(self, theta: Tensor, x: Tensor, b: BoolTensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(*, D)`, or
                a subset :math:`\theta_b`.
            x: The observation :math:`x`, with shape :math:`(*, L)`.
            b: A binary mask :math:`b`, with shape :math:`(*, D)`.

        Returns:
            The log-ratio :math:`\log r_\phi(\theta_b, x, b)`, with shape :math:`(*,)`.
        """

        if theta.shape[-1] < b.shape[-1]:
            theta, b = broadcast(theta, b, ignore=1)
            theta = theta.new_zeros(b.shape).masked_scatter(b, theta)

        theta, x, b = broadcast(theta * b, x, b * 2.0 - 1.0, ignore=1)

        return self.net(torch.cat((theta, x, b), dim=-1)).squeeze(-1)


class AMNRELoss(nn.Module):
    r"""Creates a module that calculates the cross-entropy loss for an AMNRE network.

    Given a batch of :math:`N \geq 2` pairs :math:`(\theta_i, x_i)`, the module returns

    .. math:: l = \frac{1}{2N} \sum_{i = 1}^N
        \ell(d_\phi(\theta_i \odot b_i, x_i, b_i)) +
        \ell(1 - d_\phi(\theta_{i+1} \odot b_i, x_i, b_i))

    where the binary masks :math:`b_i` are sampled from a distribution :math:`P(b)`.

    Arguments:
        estimator: A log-ratio network :math:`\log r_\phi(\theta, x, b)`.
        mask_dist: A binary mask distribution :math:`P(b)`.
    """

    def __init__(
        self,
        estimator: nn.Module,
        mask_dist: Distribution,
    ):
        super().__init__()

        self.estimator = estimator
        self.mask_dist = mask_dist

    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(N, D)`.
            x: The observation :math:`x`, with shape :math:`(N, L)`.

        Returns:
            The scalar loss :math:`l`.
        """

        theta_prime = torch.roll(theta, 1, dims=0)

        b = self.mask_dist.sample(theta.shape[:-1])

        log_r, log_r_prime = self.estimator(
            torch.stack((theta, theta_prime)),
            x,
            b,
        )

        l1 = -F.logsigmoid(log_r).mean()
        l0 = -F.logsigmoid(-log_r_prime).mean()

        return (l1 + l0) / 2


class NPE(nn.Module):
    r"""Creates a neural posterior estimation (NPE) normalizing flow.

    The principle of neural posterior estimation is to train a parametric conditional
    distribution :math:`p_\phi(\theta | x)` to approximate the posterior distribution
    :math:`p(\theta | x)`. The optimization problem is to minimize the expected
    Kullback-Leibler (KL) divergence between the two distributions for all observations
    :math:`x \sim p(x)`, that is,

    .. math::
        \arg\min_\phi & ~ \mathbb{E}_{p(x)}
            \Big[ \text{KL} \big( p(\theta|x) \parallel p_\phi(\theta | x) \big) \Big] \\
        = \arg\min_\phi & ~ \mathbb{E}_{p(x)} \, \mathbb{E}_{p(\theta | x)}
            \left[ \log \frac{p(\theta | x)}{p_\phi(\theta | x)} \right] \\
        = \arg\min_\phi & ~ \mathbb{E}_{p(\theta, x)}
            \big[ -\log p_\phi(\theta | x) \big] .

    Normalizing flows are typically used for :math:`p_\phi(\theta | x)` as they are
    differentiable parametric distributions enabling gradient-based optimization
    techniques.

    Wikipedia:
        https://wikipedia.org/wiki/Kullback-Leibler_divergence

    Arguments:
        theta_dim: The dimensionality :math:`D` of the parameter space.
        x_dim: The dimensionality :math:`L` of the observation space.
        build: The flow constructor (e.g. :class:`zuko.flows.NSF`). It takes the
            number of sample and context features as positional arguments.
        kwargs: Keyword arguments passed to the constructor.
    """

    def __init__(
        self,
        theta_dim: int,
        x_dim: int,
        build: Callable[[int, int], FlowModule] = MAF,
        **kwargs,
    ):
        super().__init__()

        self.flow = build(theta_dim, x_dim, **kwargs)

    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(*, D)`.
            x: The observation :math:`x`, with shape :math:`(*, L)`.

        Returns:
            The log-density :math:`\log p_\phi(\theta | x)`, with shape :math:`(*,)`.
        """

        theta, x = broadcast(theta, x, ignore=1)

        return self.flow(x).log_prob(theta)


class NPELoss(nn.Module):
    r"""Creates a module that calculates the negative log-likelihood loss for a NPE
    normalizing flow.

    Given a batch of :math:`N` pairs :math:`(\theta_i, x_i)`, the module returns

    .. math:: l = \frac{1}{N} \sum_{i = 1}^N -\log p_\phi(\theta_i | x_i) .

    Arguments:
        estimator: A normalizing flow :math:`p_\phi(\theta | x)`.
    """

    def __init__(self, estimator: nn.Module):
        super().__init__()

        self.estimator = estimator

    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(N, D)`.
            x: The observation :math:`x`, with shape :math:`(N, L)`.

        Returns:
            The scalar loss :math:`l`.
        """

        log_p = self.estimator(theta, x)

        return -log_p.mean()


class FMPE(nn.Module):
    r"""Creates a flow matching posterior estimation (FMPE) network.

    The principle of FMPE is to train a regression network :math:`v_\phi(\theta, x, t)`
    to approximate a vector field inducing a time-continuous normalizing flow between
    the posterior distribution :math:`p(\theta | x)` and a standard Gaussian
    distribution :math:`\mathcal{N}(0, I)`.

    After training, the normalizing flow :math:`p_\phi(\theta | x)` induced by
    :math:`v_\phi(\theta, x, t)` is used to evaluate the posterior density or generate
    samples.

    References:
        | Flow Matching for Generative Modeling (Lipman et al., 2023)
        | https://arxiv.org/abs/2210.02747

        | Flow Matching for Scalable Simulation-Based Inference (Dax et al., 2023)
        | https://arxiv.org/abs/2305.17161

    Arguments:
        theta_dim: The dimensionality :math:`D` of the parameter space.
        x_dim: The dimensionality :math:`L` of the observation space.
        freqs: The number of time embedding frequencies.
        build: The network constructor (e.g. :class:`lampe.nn.ResMLP`). It takes the
            number of input and output features as positional arguments.
        kwargs: Keyword arguments passed to the constructor.
    """

    def __init__(
        self,
        theta_dim: int,
        x_dim: int,
        freqs: int = 3,
        build: Callable[[int, int], nn.Module] = MLP,
        **kwargs,
    ):
        super().__init__()

        self.net = build(theta_dim + x_dim + 2 * freqs, theta_dim, **kwargs)

        self.register_buffer('freqs', torch.arange(1, freqs + 1) * math.pi)
        self.register_buffer('zeros', torch.zeros(theta_dim))
        self.register_buffer('ones', torch.ones(theta_dim))

    def forward(self, theta: Tensor, x: Tensor, t: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(*, D)`.
            x: The observation :math:`x`, with shape :math:`(*, L)`.
            t: The time :math:`t`, with shape :math:`(*,).`

        Returns:
            The vector field :math:`v_\phi(\theta, x, t)`, with shape :math:`(*, D)`.
        """

        t = self.freqs * t[..., None]
        t = torch.cat((t.cos(), t.sin()), dim=-1)

        theta, x, t = broadcast(theta, x, t, ignore=1)

        return self.net(torch.cat((theta, x, t), dim=-1))

    def flow(self, x: Tensor) -> Distribution:
        r"""
        Arguments:
            x: The observation :math:`x`, with shape :math:`(*, L)`.

        Returns:
            The normalizing flow :math:`p_\phi(\theta | x)`.
        """

        return NormalizingFlow(
            transform=FreeFormJacobianTransform(
                f=lambda t, theta: self(theta, x, t),
                t0=x.new_tensor(0.0),
                t1=x.new_tensor(1.0),
                phi=(x, *self.parameters()),
            ),
            base=DiagNormal(self.zeros, self.ones).expand(x.shape[:-1]),
        )


class FMPELoss(nn.Module):
    r"""Creates a module that calculates the flow matching loss for a FMPE regressor.

    Given a batch of :math:`N` pairs :math:`(\theta_i, x_i)`, the module returns

    .. math:: l = \frac{1}{N} \sum_{i = 1}^N\|
            v_\phi((1 - t_i) \theta_i + (t_i + \eta) \epsilon_i, x_i, t_i)
            - (\epsilon_i - \theta_i)
        \|_2^2

    where :math:`t_i \sim \mathcal{U}(0, 1)` and :math:`\epsilon_i \sim \mathcal{N}(0, I)`.

    Arguments:
        estimator: A regression network :math:`v_\phi(\theta, x, t)`.
    """

    def __init__(self, estimator: nn.Module, eta: float = 1e-3):
        super().__init__()

        self.estimator = estimator
        self.eta = eta

    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""
        Arguments:
            theta: The parameters :math:`\theta`, with shape :math:`(N, D)`.
            x: The observation :math:`x`, with shape :math:`(N, L)`.

        Returns:
            The scalar loss :math:`l`.
        """

        t = torch.rand(theta.shape[:-1], dtype=theta.dtype, device=theta.device)
        t_ = t[..., None]

        eps = torch.randn_like(theta)
        theta_prime = (1 - t_) * theta + (t_ + self.eta) * eps
        v = eps - theta

        return (self.estimator(theta_prime, x, t) - v).square().mean()


class MetropolisHastings(object):
    r"""Creates a batched Metropolis-Hastings sampler.

    Metropolis-Hastings is a Markov chain Monte Carlo (MCMC) sampling algorithm used to
    sample from intractable distributions :math:`p(x)` whose density is proportional to
    a tractable function :math:`f(x)`, with :math:`x \in \mathcal{X}`. The algorithm
    consists in repeating the following routine for :math:`t = 1` to :math:`T`, where
    :math:`x_0` is the initial sample and :math:`q(x' | x)` is a pre-defined transition
    distribution.

    .. math::
        1. ~ & x' \sim q(x' | x_{t-1}) \\
        2. ~ & \alpha \gets \frac{f(x')}{f(x_{t-1})}
            \frac{q(x_{t-1} | x')}{q(x' | x_{t-1})} \\
        3. ~ & u \sim \mathcal{U}(0, 1) \\
        4. ~ & x_t \gets \begin{cases}
            x' & \text{if } u \leq \alpha \\
            x_{t-1} & \text{otherwise}
        \end{cases}

    Asymptotically, i.e. when :math:`T \to \infty`, the distribution of samples
    :math:`x_t` is guaranteed to converge towards :math:`p(x)`. In this implementation,
    a Gaussian transition :math:`q(x' | x) = \mathcal{N}(x'; x, \Sigma)` is used, which
    can be modified by sub-classing :class:`MetropolisHastings`.

    Wikipedia:
        https://wikipedia.org/wiki/Metropolis-Hastings_algorithm

    Arguments:
        x_0: A batch of initial points :math:`x_0`, with shape :math:`(*, L)`.
        f: A function :math:`f(x)` proportional to a density function :math:`p(x)`.
        log_f: The logarithm :math:`\log f(x)` of a function proportional
            to :math:`p(x)`.
        sigma: The standard deviation of the Gaussian transition.
            Either a scalar or a vector.

    Example:
        >>> x_0 = torch.randn(128, 7)
        >>> log_f = lambda x: -(x**2).sum(dim=-1) / 2
        >>> sampler = MetropolisHastings(x_0, log_f=log_f, sigma=0.5)
        >>> samples = [x for x in sampler(256, burn=128, step=4)]
        >>> samples = torch.stack(samples)
        >>> samples.shape
        torch.Size([32, 128, 7])
    """

    def __init__(
        self,
        x_0: Tensor,
        f: Callable[[Tensor], Tensor] = None,
        log_f: Callable[[Tensor], Tensor] = None,
        sigma: Union[float, Tensor] = 1.0,
    ):
        super().__init__()

        self.x_0 = x_0

        assert f is not None or log_f is not None, \
            "Either 'f' or 'log_f' has to be provided."

        if f is None:
            self.log_f = log_f
        else:
            self.log_f = lambda x: f(x).log()

        self.sigma = sigma

    def q(self, x: Tensor) -> Distribution:
        return DiagNormal(x, torch.ones_like(x) * self.sigma)

    @property
    def symmetric(self) -> bool:
        return True

    def __iter__(self) -> Iterator[Tensor]:
        x = self.x_0

        # log f(x)
        log_f_x = self.log_f(x)

        while True:
            # y ~ q(y | x)
            y = self.q(x).sample()

            # log f(y)
            log_f_y = self.log_f(y)

            #     f(y)   q(x | y)
            # a = ---- * --------
            #     f(x)   q(y | x)
            log_a = log_f_y - log_f_x

            if not self.symmetric:
                log_a = log_a + self.q(y).log_prob(x) - self.q(x).log_prob(y)

            a = log_a.exp()

            # u in [0; 1]
            u = torch.rand(a.shape).to(a)

            # if u < a, x <- y
            # else x <- x
            mask = u < a

            x = torch.where(mask.unsqueeze(-1), y, x)
            log_f_x = torch.where(mask, log_f_y, log_f_x)

            yield x

    def __call__(self, stop: int, burn: int = 0, step: int = 1) -> Iterable[Tensor]:
        return islice(self, burn, stop, step)



r"""Masking helpers."""

import torch
import torch.nn as nn

from torch import Tensor, BoolTensor, Size
from torch.distributions import Distribution, Bernoulli, Independent
from typing import *


def mask2str(b: BoolTensor) -> str:
    r"""Represents a binary mask as a string.

    Arguments:
        b: A binary mask :math:`b`, with shape :math:`(D,)`.

    Example:
        >>> b = torch.tensor([True, True, False, True, False])
        >>> mask2str(b)
        '11010'
    """

    return ''.join('1' if bit else '0' for bit in b)


def str2mask(string: str) -> BoolTensor:
    r"""Parses the string representation of a binary mask into a tensor.

    Arguments:
        string: A binary mask string representation.

    Example:
        >>> str2mask('11010')
        tensor([True, True, False, True, False])
    """

    return torch.tensor([char == '1' for char in string])


class BernoulliMask(Independent):
    r"""Creates a distribution :math:`P(b)` over all binary masks :math:`b` in the
    hypercube :math:`\{0, 1\}^D` such that each bit :math:`b_i` has a probability
    :math:`p_i` of being positive.

    .. math:: P(b) = \prod^D_{i = 1} p_i^{b_i} (1 - p_i)^{1 - b_i}

    Arguments:
        p: The probability vector :math:`p`, with shape :math:`(*, D)`.

    Example:
        >>> d = BernoulliMask(torch.tensor([0.4, 0.1, 0.9]))
        >>> d.sample()
        tensor([True, False, True])
    """

    has_rsample = False

    def __init__(self, p: Tensor):
        super().__init__(Bernoulli(p), 1)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(D={self.event_shape.numel()})'

    def log_prob(self, b: BoolTensor) -> Tensor:
        return super().log_prob(b.float())

    def sample(self, shape: Size = ()) -> BoolTensor:
        return super().sample(shape).bool()


class SelectionMask(Distribution):
    r"""Creates a mask distribution :math:`P(b)`, uniform over a selection of
    binary masks :math:`\mathcal{B} \subseteq \{0, 1\}^D`.

    .. math:: P(b) = \begin{cases}
            \frac{1}{|\mathcal{B}|} & \text{if } b \in \mathcal{B} \\
            0 & \text{otherwise}
        \end{cases}

    Arguments:
        selection: A binary mask selection :math:`\mathcal{B}`, with shape :math:`(N, D)`.

    Example:
        >>> selection = torch.tensor([
        ...     [True, False, False],
        ...     [False, True, False],
        ...     [False, False, True],
        ... ])
        >>> d = SelectionMask(selection)
        >>> d.sample()
        tensor([False, True, False])
    """

    def __init__(self, selection: BoolTensor):
        super().__init__(event_shape=selection.shape[-1:])

        self.selection = selection

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(D={self.event_shape.numel()})'

    def log_prob(self, b: BoolTensor) -> Tensor:
        match = torch.all(b[..., None, :] == self.selection, dim=-1)
        prob = match.float().mean(dim=-1)
        return prob.log()

    def sample(self, shape: Size = ()) -> BoolTensor:
        index = torch.randint(len(self.selection), shape)
        return self.selection[index]



r"""Tests for the lampe.diagnostics module."""

import pytest
import torch

from lampe.diagnostics import *
from zuko.distributions import Independent, Normal, Truncated


def test_expected_coverage_mc():
    posterior = lambda x: Independent(Normal(0, 1 + x**2), 1)
    x = torch.randn(1024, 2)
    theta = posterior(x).sample()
    pairs = list(zip(theta, x))

    # Exact
    estimator = posterior
    levels, coverages = expected_coverage_mc(estimator, pairs, n=1024)

    assert torch.allclose(levels, coverages, atol=1e-1)

    # Conservative
    estimator = lambda x: Independent(Normal(0, 2 + x**2), 1)
    levels, coverages = expected_coverage_mc(estimator, pairs, n=1024)

    assert (coverages > levels).float().mean() > 0.9

    # Overconfident
    estimator = lambda x: Independent(Normal(0, 0.5 + x**2), 1)
    levels, coverages = expected_coverage_mc(estimator, pairs, n=1024)

    assert (coverages < levels).float().mean() > 0.9


def test_expected_coverage_ni():
    posterior = lambda x: Independent(Truncated(Normal(0, 1 + x**2), -3, 3), 1)
    x = torch.randn(1024, 2)
    theta = posterior(x).sample()
    pairs = list(zip(theta, x))
    domain = (-3 * torch.ones(2), 3 * torch.ones(2))

    # Exact
    estimator = lambda theta, x: posterior(x).log_prob(theta)
    levels, coverages = expected_coverage_ni(estimator, pairs, domain, bins=128)

    assert torch.allclose(levels, coverages, atol=1e-1)

    # Conservative
    estimator = lambda theta, x: Independent(
        Truncated(Normal(0, 2 + x**2), -3, 3), 1
    ).log_prob(theta)
    levels, coverages = expected_coverage_ni(estimator, pairs, domain, bins=128)

    assert (coverages > levels).float().mean() > 0.9

    # Overconfident
    estimator = lambda theta, x: Independent(
        Truncated(Normal(0, 0.5 + x**2), -3, 3), 1
    ).log_prob(theta)
    levels, coverages = expected_coverage_ni(estimator, pairs, domain, bins=128)

    assert (coverages < levels).float().mean() > 0.9



r"""Datasets and data loaders."""

__all__ = ['JointLoader', 'JointDataset', 'H5Dataset']

import h5py
import numpy as np
import random
import torch

from bisect import bisect
from numpy import ndarray as Array
from pathlib import Path
from torch import Tensor, Size
from torch.distributions import Distribution
from torch.utils.data import DataLoader, Dataset, IterableDataset
from tqdm import tqdm
from typing import *


class IterableJointDataset(IterableDataset):
    r"""Creates an iterable dataset of batched pairs :math:`(\theta, x)`."""

    def __init__(
        self,
        prior: Distribution,
        simulator: Callable,
        batch_shape: Size = (),
        numpy: bool = False,
    ):
        super().__init__()

        self.prior = prior
        self.simulator = simulator
        self.batch_shape = batch_shape
        self.numpy = numpy

    def __iter__(self) -> Iterator[Tuple[Tensor, Tensor]]:
        while True:
            theta = self.prior.sample(self.batch_shape)

            if self.numpy:
                x = self.simulator(theta.detach().cpu().numpy().astype(np.double))
                x = torch.from_numpy(x).to(theta)
            else:
                x = self.simulator(theta)

            yield theta, x


class JointLoader(DataLoader):
    r"""Creates an infinite data loader of batched pairs :math:`(\theta, x)` generated
    by a prior distribution :math:`p(\theta)` and a simulator.

    The simulator is a stochastic function taking (a vector of) parameters
    :math:`\theta`, in the form of a NumPy array or a PyTorch tensor, as input and
    returning an observation :math:`x` as output, which implicitly defines a
    likelihood distribution :math:`p(x | \theta)`. Together with the prior, they form
    a joint distribution :math:`p(\theta, x) = p(\theta) p(x | \theta)` from which
    pairs :math:`(\theta, x)` are independently drawn.

    Arguments:
        prior: A prior distribution :math:`p(\theta)`.
        simulator: A callable simulator.
        batch_size: The batch size of the generated pairs.
        vectorized: Whether the simulator accepts batched inputs or not.
        numpy: Whether the simulator requires NumPy or PyTorch inputs.
        kwargs: Keyword arguments passed to :class:`torch.utils.data.DataLoader`.

    Example:
        >>> loader = JointLoader(prior, simulator, numpy=True, num_workers=4)
        >>> for theta, x in loader:
        ...     theta, x = theta.cuda(), x.cuda()
        ...     something(theta, x)
    """

    def __init__(
        self,
        prior: Distribution,
        simulator: Callable,
        batch_size: int = 2**8,  # 256
        vectorized: bool = False,
        numpy: bool = False,
        **kwargs,
    ):
        super().__init__(
            IterableJointDataset(
                prior,
                simulator,
                batch_shape=(batch_size,) if vectorized else (),
                numpy=numpy,
            ),
            batch_size=None if vectorized else batch_size,
            **kwargs,
        )


class JointDataset(Dataset):
    r"""Creates an in-memory dataset of pairs :math:`(\theta, x)`.

    :class:`JointDataset` supports indexing and slicing, but also implements a custom
    :meth:`__iter__` method which supports batching and shuffling.

    Arguments:
        theta: A tensor of parameters :math:`\theta`.
        x: A tensor of observations :math:`x`.
        batch_size: The size of the batches.
        shuffle: Whether the pairs are shuffled or not when iterating.

    Example:
        >>> dataset = JointDataset(theta, x, batch_size=256, shuffle=True)
        >>> theta, x = dataset[42:69]
        >>> theta.shape
        torch.Size([27, 5])
        >>> for theta, x in dataset:
        ...     theta, x = theta.cuda(), x.cuda()
        ...     something(theta, x)
    """

    def __init__(
        self,
        theta: Tensor,
        x: Tensor,
        batch_size: int = None,
        shuffle: bool = False,
    ):
        super().__init__()

        assert len(theta) == len(x)

        self.theta = torch.as_tensor(theta)
        self.x = torch.as_tensor(x)

        self.batch_size = batch_size
        self.shuffle = shuffle

    def __len__(self) -> int:
        return len(self.theta)

    def __getitem__(self, i: Union[int, slice]) -> Tuple[Tensor, Tensor]:
        return self.theta[i], self.x[i]

    def __iter__(self) -> Iterator[Tuple[Tensor, Tensor]]:
        if self.shuffle:
            order = torch.randperm(len(self))

            if self.batch_size is None:
                return (self[i] for i in order)
            else:
                return (self[i] for i in order.split(self.batch_size))
        else:
            if self.batch_size is None:
                return zip(self.theta, self.x)
            else:
                return zip(
                    self.theta.split(self.batch_size),
                    self.x.split(self.batch_size),
                )


class H5Dataset(IterableDataset):
    r"""Creates an iterable dataset of pairs :math:`(\theta, x)` from an HDF5 file.

    As it can be slow to load pairs from disk one by one, :class:`H5Dataset` implements
    a custom :meth:`__iter__` method that loads several contiguous chunks of pairs at
    once and shuffles their concatenation before yielding the pairs.

    :class:`H5Dataset` also implements the :meth:`__len__` and :meth:`__getitem__`
    methods for convenience.

    Important:
        When using :class:`H5Dataset` with :class:`torch.utils.data.DataLoader`, it is
        recommended to disable the loader's automatic batching and provide the batch
        size to the dataset, as it significantly improves loading performances.
        In fact, unless using several workers or memory pinning, it is not even
        necessary to wrap the dataset in a :class:`torch.utils.data.DataLoader`.

    Arguments:
        file: An HDF5 file containing pairs :math:`(\theta, x)`.
        batch_size: The size of the batches.
        chunk_size: The size of the contiguous chunks.
        chunk_step: The number of chunks loaded at once.
        shuffle: Whether the pairs are shuffled or not when iterating.

    Example:
        >>> dataset = H5Dataset('data.h5', batch_size=256, shuffle=True)
        >>> theta, x = dataset[0]
        >>> theta
        tensor([-0.1215, -1.3641,  0.7233, -1.2150, -1.9263])
        >>> for theta, x in dataset:
        ...     theta, x = theta.cuda(), x.cuda()
        ...     something(theta, x)
    """

    def __init__(
        self,
        file: Union[str, Path],
        batch_size: int = None,
        chunk_size: int = 2**8,  # 256
        chunk_step: str = 2**8,  # 256
        shuffle: bool = False,
    ):
        super().__init__()

        self.file = h5py.File(file, mode='r')

        self.batch_size = batch_size
        self.chunk_size = chunk_size
        self.chunk_step = chunk_step
        self.shuffle = shuffle

    def __len__(self) -> int:
        return len(self.file['theta'])

    def __getitem__(self, i: Union[int, slice]) -> Tuple[Tensor, Tensor]:
        theta, x = self.file['theta'][i], self.file['x'][i]

        return torch.from_numpy(theta), torch.from_numpy(x)

    def __iter__(self) -> Iterator[Tuple[Tensor, Tensor]]:
        chunks = torch.tensor([
            (i, i + self.chunk_size)
            for i in range(0, len(self), self.chunk_size)
        ])

        if self.shuffle:
            order = torch.randperm(len(chunks))
            chunks = chunks[order]

        for slices in chunks.split(self.chunk_step):
            # Merge contiguous slices
            slices = sorted(slices.tolist())
            stack = []

            for s in slices:
                if stack and stack[-1][-1] == s[0]:
                    stack[-1][-1] = s[-1]
                else:
                    stack.append(s)

            # Load
            theta = np.concatenate([self.file['theta'][i:j] for i, j in stack])
            x = np.concatenate([self.file['x'][i:j] for i, j in stack])

            theta, x = torch.from_numpy(theta), torch.from_numpy(x)

            # Shuffle
            if self.shuffle:
                order = torch.randperm(len(theta))
                theta, x = theta[order], x[order]

            # Batch
            if self.batch_size is None:
                yield from zip(theta, x)
            else:
                yield from zip(
                    theta.split(self.batch_size),
                    x.split(self.batch_size),
                )

    def to_memory(self) -> JointDataset:
        r"""Loads all pairs in memory and returns them as a :class:`JointDataset`.

        Example:
            >>> dataset = H5Dataset('data.h5').to_memory()
        """

        return JointDataset(
            self.file['theta'][:],
            self.file['x'][:],
            batch_size=self.batch_size,
            shuffle=self.shuffle,
        )

    @staticmethod
    def store(
        pairs: Iterable[Tuple[Array, Array]],
        file: Union[str, Path],
        size: int,
        overwrite: bool = False,
        dtype: np.dtype = np.float32,
        **meta,
    ) -> None:
        r"""Creates an HDF5 file containing pairs :math:`(\theta, x)`.

        The sets of parameters :math:`\theta` are stored in a collection named
        :py:`'theta'` and the observations in a collection named :py:`'x'`.

        Arguments:
            pairs: An iterable over batched pairs :math:`(\theta, x)`.
            file: An HDF5 filename to store pairs in.
            size: The number of pairs to store.
            overwrite: Whether to overwrite existing files or not. If :py:`False`
                and the file already exists, the function raises an error.
            dtype: The data type to store pairs in.
            meta: Metadata to store in the file.

        Example:
            >>> loader = JointLoader(prior, simulator, batch_size=16)
            >>> H5Dataset.store(loader, 'data.h5', 4096)
            100%|██████████| 4096/4096 [01:35<00:00, 42.69pair/s]
        """

        # Pairs
        pairs = iter(pairs)

        # File
        file = Path(file)
        file.parent.mkdir(parents=True, exist_ok=True)

        with h5py.File(file, 'w' if overwrite else 'w-') as f:
            ## Attributes
            f.attrs.update(meta)

            ## Datasets
            theta, x = map(np.asarray, next(pairs))

            f.create_dataset('theta', (size,) + theta.shape[1:], dtype=dtype)
            f.create_dataset('x', (size,) + x.shape[1:], dtype=dtype)

            ## Store
            with tqdm(total=size, unit='pair') as tq:
                i = 0

                while True:
                    j = min(i + theta.shape[0], size)

                    f['theta'][i:j] = theta[: j - i]
                    f['x'][i:j] = x[: j - i]

                    tq.update(j - i)

                    if j < size:
                        i = j

                        try:
                            theta, x = map(np.asarray, next(pairs))
                        except StopIteration:
                            break
                    else:
                        break



r"""Tests for the lampe.masks module."""

import pytest
import torch

from lampe.masks import *


def test_mask2str():
    b = torch.tensor([True, True, False, True, False])

    assert mask2str(b) == '11010'
    assert (str2mask('11010') == b).all()


def test_BernouilliMask():
    p = torch.rand(3)
    d = BernoulliMask(p)

    assert d.event_shape == (3,)

    b = d.sample()

    assert b.shape == (3,)
    assert b.dtype == torch.bool

    b = d.sample((5, 4))

    assert b.shape == (5, 4, 3)

    lp = d.log_prob(b)

    assert lp.shape == (5, 4)


def test_SelectionMask():
    selection = torch.tensor([
        [True, False, False],
        [False, True, False],
        [False, False, True],
    ])
    d = SelectionMask(selection)

    assert d.event_shape == (3,)

    b = d.sample()

    assert b.shape == (3,)
    assert b.dtype == torch.bool

    b = d.sample((5, 4))

    assert b.shape == (5, 4, 3)

    lp = d.log_prob(b)

    assert lp.shape == (5, 4)



r"""Tests for the lampe.nn module."""

import pytest
import torch

from lampe.nn import *
from torch import randn


def test_ResMLP():
    net = ResMLP(3, 5)

    # Non-batched
    x = randn(3)
    y = net(x)

    assert y.shape == (5,)
    assert y.requires_grad

    # Batched
    x = randn(256, 3)
    y = net(x)

    assert y.shape == (256, 5)



r"""Plotting helpers."""

__all__ = ['nice_rc', 'corner', 'mark_point', 'coverage_plot']

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from numpy import ndarray as Array
from typing import *


def nice_rc(latex: bool = False) -> Dict[str, Any]:
    r"""Returns a dictionary of runtime configuration (rc) settings for nicer
    :mod:`matplotlib` plots. The settings include 12pt font size, higher DPI,
    tight layout, transparent background, etc.

    Arguments:
        latex: Whether to use LaTeX typesetting or not.

    Example:
        >>> plt.rcParams.update(nice_rc())
    """

    rc = {
        'axes.axisbelow': True,
        'axes.linewidth': 0.8,
        'figure.dpi': 150,
        'figure.figsize': (6.4, 4.8),
        'font.size': 12.0,
        'legend.fontsize': 'x-small',
        'lines.linewidth': 1.0,
        'lines.markersize': 3.0,
        'savefig.transparent': True,
        'xtick.labelsize': 'x-small',
        'xtick.major.width': 0.8,
        'ytick.labelsize': 'x-small',
        'ytick.major.width': 0.8,
    }

    if latex:
        rc.update({
            'font.family': ['serif'],
            'font.serif': ['Computer Modern'],
            'text.usetex': True,
        })

    return rc


class LinearAlphaColormap(mpl.colors.LinearSegmentedColormap):
    r"""Linear segmented transparency colormap.

    Arguments:
        color: A color.
        levels: A sequence of levels dividing the domain into segments.
        alpha: The transparency range.
        name: A name for the colormap.
    """

    def __new__(
        self,
        color: Union[str, tuple],
        levels: Array = None,
        alpha: Tuple[float, float] = (0.0, 1.0),
        name: str = None,
    ):
        if name is None:
            if type(color) is str:
                name = f'alpha_{color}'
            else:
                name = f'alpha_{hash(color)}'

        if levels is None:
            levels = [0.0, 1.0]

        levels = np.asarray(levels)
        levels = (levels - levels.min()) / (levels.max() - levels.min())
        alpha = np.linspace(*alpha, len(levels))

        rgb = mpl.colors.to_rgb(color)
        colors = sorted((l, rgb + (a,)) for l, a in zip(levels, alpha))

        return mpl.colors.LinearSegmentedColormap.from_list(
            name=name,
            colors=colors,
        )


def gaussian_blur(img: Array, sigma: float = 1.0) -> Array:
    r"""Applies a Gaussian blur to an image.

    Arguments:
        img: An image array.
        sigma: The standard deviation of the Gaussian kernel.

    Returns:
        The blurred image.

    Example:
        >>> img = np.random.rand(128, 128)
        >>> gaussian_blur(img, sigma=2.0)
        array([...])
    """

    size = 2 * int(3 * sigma) + 1

    k = np.arange(size) - size / 2
    k = np.exp(-(k**2) / (2 * sigma**2))
    k = k / np.sum(k)

    smooth = lambda x: np.convolve(x, k, mode='same')

    for i in range(len(img.shape)):
        img = np.apply_along_axis(smooth, i, img)

    return img


def credible_levels(hist: Array, creds: Array) -> Array:
    r"""Returns the levels of credibility region contours.

    Arguments:
        hist: An histogram.
        creds: The region credibilities.
    """

    x = np.sort(hist, axis=None)[::-1]
    cdf = np.cumsum(x)
    idx = np.searchsorted(cdf, creds * cdf[-1])

    return x[idx]


def corner(
    data: Array,
    weights: Array = None,
    domain: Tuple[Array, Array] = None,
    bins: Union[int, Sequence[int]] = 64,
    creds: Sequence[float] = (0.6827, 0.9545, 0.9973),
    color: Union[str, tuple] = None,
    alpha: Tuple[float, float] = (0.0, 0.5),
    legend: str = None,
    labels: Sequence[str] = None,
    smooth: float = 0,
    figure: mpl.figure.Figure = None,
    **kwargs,
) -> mpl.figure.Figure:
    r"""Displays each 1 or 2-d projection of multi-dimensional data, as a triangular
    matrix of histograms, known as corner plot. For 2-d histograms, highest density
    credibility regions are delimited.

    Arguments:
        data: Multi-dimensional data, either as a table or as a matrix of histograms.
        weights: The importance weights of the data samples.
        domain: A pair of lower and upper domain bounds. If :py:`None`, inferred from data.
        bins: The number(s) of bins per dimension.
        creds: The region credibilities (in :math:`[0, 1]`) to delimit.
        color: The color of histograms. If :py:`None`, use the color cycler instead.
        alpha: The transparency range.
        legend: A legend.
        labels: The dimension labels.
        smooth: The standard deviation of the smoothing kernels.
        figure: A corner plot over which to draw the new one.
        kwargs: Keyword arguments passed to :func:`matplotlib.pyplot.subplots`.

    Returns:
        The figure instance for the corner plot.

    Example:
        >>> data = np.random.randn(2**16, 3)
        >>> labels = [r'$\alpha$', r'$\beta$', r'$\gamma$']
        >>> figure = corner(data, bins=32, labels=labels, figsize=(4.8, 4.8))

    .. image:: ../images/corner.png
        :align: center
        :width: 600
    """

    # Histograms
    if np.asarray(data[0][0]).ndim == 0:
        data = np.asarray(data)
        D = data.shape[-1]

        if type(bins) is int:
            bins = [bins] * D

        if domain is None:
            lower, upper = data.min(axis=0), data.max(axis=0)
        else:
            lower, upper = map(np.asarray, domain)

        bins = [
            np.histogram_bin_edges(data, bins[i], range=(lower[i], upper[i]))
            for i in range(D)
        ]

        hists = np.ndarray((D, D), dtype=object)

        for i in range(D):
            for j in range(i + 1):
                if i == j:
                    hist, _ = np.histogram(
                        data[..., i],
                        bins=bins[i],
                        density=True,
                        weights=weights,
                    )
                else:
                    hist, _, _ = np.histogram2d(
                        data[..., i],
                        data[..., j],
                        bins=(bins[i], bins[j]),
                        density=True,
                        weights=weights,
                    )

                hists[i, j] = hist
    else:
        data = np.asarray(data, dtype=object)
        D = len(data)

        if domain is None:
            lower, upper = np.zeros(D), np.ones(D)
        else:
            lower, upper = map(np.asarray, domain)

        bins = [None] * D
        for i in range(D):
            if data[i, i] is not None:
                bins[i] = np.linspace(lower[i], upper[i], len(data[i, i]) + 1)

        hists = data

    # Figure
    if figure is None:
        kwargs.setdefault('figsize', (6.4, 6.4))

        figure, axes = plt.subplots(
            D,
            D,
            squeeze=False,
            sharex='col',
            gridspec_kw={'wspace': 0.0, 'hspace': 0.0},
            **kwargs,
        )
        new = True
    else:
        axes = np.asarray(figure.axes).reshape(D, D)
        new = False

    # Legend

    ## Quantiles
    creds = np.sort(np.asarray(creds))[::-1]
    creds = np.append(creds, 0)

    if new:
        cmap = LinearAlphaColormap('black', levels=creds, alpha=alpha)

        levels = (creds - creds.min()) / (creds.max() - creds.min())
        levels = (levels[:-1] + levels[1:]) / 2

        for c, l in zip(creds[:-1], levels):
            axes[0, -1].plot(
                [],
                [],
                color=cmap(l),
                linewidth=6,
                solid_capstyle='butt',
                label=r'${:.1f}\,\%$'.format(c * 100),
            )

    ## Color
    lines = axes[0, -1].plot([], [], color=color, solid_capstyle='butt', label=legend)
    color = lines[-1].get_color()

    axes[0, -1].legend(loc='upper right', frameon=False)

    # Plot
    for i in range(D):
        for j in range(D):
            ax = axes[i, j]
            hist = hists[i, j]

            if j > i:
                ax.axis('off')
                continue

            if hist is None:
                continue

            if smooth > 0:
                hist = gaussian_blur(hist, smooth)

            ## Draw
            x, y = bins[j], bins[i]
            x = (x[1:] + x[:-1]) / 2
            y = (y[1:] + y[:-1]) / 2

            if i == j:
                ax.plot(x, hist, color=color)

                _, top = ax.get_ylim()
                top = max(top, hist.max() * 1.0625)

                ax.set_xlim(left=bins[i][0], right=bins[i][-1])
                ax.set_ylim(bottom=0.0, top=top)
            else:
                levels = np.unique(credible_levels(hist, creds))

                cf = ax.contourf(
                    x,
                    y,
                    hist,
                    levels=levels,
                    cmap=LinearAlphaColormap(color, levels, alpha=alpha),
                )
                ax.contour(cf, colors=color)

                if j > 0:
                    ax.sharey(axes[i, j - 1])
                else:
                    ax.set_ylim(bottom=bins[i][0], top=bins[i][-1])

            ## Ticks
            if i == D - 1:
                ax.xaxis.set_major_locator(mpl.ticker.MaxNLocator(3, prune='both'))
                plt.setp(
                    ax.get_xticklabels(),
                    rotation=45.0,
                    horizontalalignment='right',
                    rotation_mode='anchor',
                )
            else:
                ax.xaxis.set_ticks_position('none')

            if i == j:
                ax.set_yticks([])
            elif j == 0:
                ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(3, prune='both'))
            else:
                ax.yaxis.set_ticks_position('none')

            ## Labels
            if labels is not None:
                if i == D - 1:
                    ax.set_xlabel(labels[j])

                if j == 0 and i != j:
                    ax.set_ylabel(labels[i])

            ax.label_outer()

    figure.align_labels()
    figure.tight_layout(pad=0.5)

    return figure


def mark_point(
    figure: mpl.figure.Figure,
    point: Array,
    color: Union[str, tuple] = 'black',
    linestyle: str = 'dashed',
    marker: str = 's',  # square
    legend: str = None,
) -> None:
    r"""Marks a point on the histograms of a corner plot.

    Arguments:
        figure: The corner plot figure (see :func:`corner`).
        point: The point to mark.
        color: The color of the lines and markers.
        linestyle: The style of the lines.
        marker: The shape of the markers.
        legend: A legend.

    Example:
        >>> mark_point(figure, [0.5, 0.3, -0.7], color='black')

    .. image:: ../images/corner_marked.png
        :align: center
        :width: 600
    """

    D = len(point)
    axes = np.asarray(figure.axes).reshape(D, D)

    # Legend
    if legend is not None:
        axes[0, -1].plot(
            [],
            [],
            color=color,
            linestyle=linestyle,
            marker=marker,
            solid_capstyle='butt',
            label=legend,
        )

        axes[0, -1].legend(loc='upper right', frameon=False)

    # Plot
    for i in range(D):
        for j in range(i + 1):
            ax = axes[i, j]
            ax.axvline(point[j], color=color, linestyle=linestyle)

            if i != j:
                ax.axhline(point[i], color=color, linestyle=linestyle)
                ax.plot(point[j], point[i], color=color, marker=marker)


def coverage_plot(
    levels: Array,
    coverages: Array,
    color: Union[str, tuple] = None,
    legend: str = None,
    figure: mpl.figure.Figure = None,
    **kwargs,
) -> mpl.figure.Figure:
    r"""Plots the expected coverage at various credible levels.

    Arguments:
        levels: A vector of increasing credible levels.
        coverages: A vector of corresponding expected coverages.
        legend: A legend.
        figure: Another coverage plot over which to draw the new one.
        kwargs: Keyword arguments passed to :func:`matplotlib.pyplot.subplots`.

    Returns:
        The figure instance for the coverage plot.

    Example:
        >>> levels = np.linspace(0, 1, 512) ** 2
        >>> coverages = np.linspace(0, 1, 512)
        >>> figure = coverage_plot(levels, coverages)

    .. image:: ../images/coverage.png
        :align: center
        :width: 400
    """

    levels, coverages = np.asarray(levels), np.asarray(coverages)

    # Figure
    if figure is None:
        kwargs.setdefault('figsize', (3.2, 3.2))

        figure, ax = plt.subplots(**kwargs)
        new = True
    else:
        ax = figure.axes[0]
        new = False

    # Plot
    if new:
        ax.plot([0, 1], [0, 1], color='k', linestyle='--')

    ax.plot(levels, coverages, color=color, label=legend)

    ax.grid(visible=True)
    ax.set_xlabel(r'Credible level')
    ax.set_ylabel(r'Expected coverage')

    if legend is not None:
        ax.legend(loc='upper left')

    figure.tight_layout(pad=0.5)

    return figure



r"""Tests for the lampe.inference module."""

import pytest
import torch

from lampe.inference import *
from lampe.masks import BernoulliMask
from torch import randn


def test_NRE():
    estimator = NRE(3, 5)

    # Non-batched
    theta, x = randn(3), randn(5)
    log_r = estimator(theta, x)

    assert log_r.shape == ()
    assert log_r.requires_grad

    # Batched
    theta, x = randn(256, 3), randn(256, 5)
    log_r = estimator(theta, x)

    assert log_r.shape == (256,)

    # Mixed
    theta, x = randn(256, 3), randn(5)
    log_r = estimator(theta, x)

    assert log_r.shape == (256,)

    # Losses
    losses = [
        NRELoss(estimator),
        BNRELoss(estimator),
        CNRELoss(estimator),
        BCNRELoss(estimator),
    ]

    for loss in losses:
        theta, x = randn(256, 3), randn(256, 5)

        l = loss(theta, x)

        assert l.shape == (), loss
        assert l.requires_grad, loss


def test_AMNRE():
    estimator = AMNRE(3, 5)

    # Non-batched
    theta, x, b = randn(3), randn(5), randn(3) < 0
    log_r = estimator(theta, x, b)

    assert log_r.shape == ()
    assert log_r.requires_grad

    # Batched
    theta, x, b = randn(256, 3), randn(256, 5), randn(256, 3) < 0
    log_r = estimator(theta, x, b)

    assert log_r.shape == (256,)

    grad = torch.autograd.functional.jacobian(lambda theta: estimator(theta, x, b).sum(), theta)

    assert (grad[~b] == 0).all()

    # Mixed
    theta, x, b = randn(256, 3), randn(5), randn(2, 1, 3)
    log_r = estimator(theta, x, b)

    assert log_r.shape == (2, 256)

    # Subset
    theta, x, b = randn(256, 3), randn(256, 5), torch.tensor([True, False, True])

    log_r1 = estimator(theta, x, b)
    log_r2 = estimator(theta[..., b], x, b)

    assert torch.allclose(log_r1, log_r2)


def test_AMNRELoss():
    estimator = AMNRE(3, 5)
    mask_dist = BernoulliMask(0.5 * torch.ones(3))
    loss = AMNRELoss(estimator, mask_dist)

    theta, x = randn(256, 3), randn(256, 5)

    l = loss(theta, x)

    assert l.shape == ()
    assert l.requires_grad


def test_NPE():
    estimator = NPE(3, 5)

    # Non-batched
    theta, x = randn(3), randn(5)
    log_p = estimator(theta, x)

    assert log_p.shape == ()
    assert log_p.requires_grad

    # Batched
    theta, x = randn(256, 3), randn(256, 5)
    log_p = estimator(theta, x)

    assert log_p.shape == (256,)

    # Mixed
    theta, x = randn(256, 3), randn(5)
    log_p = estimator(theta, x)

    assert log_p.shape == (256,)

    # Sample
    x = randn(32, 5)
    theta = estimator.flow(x).sample((8,))

    assert theta.shape == (8, 32, 3)


def test_NPELoss():
    estimator = NPE(3, 5)
    loss = NPELoss(estimator)

    theta, x = randn(256, 3), randn(256, 5)

    l = loss(theta, x)

    assert l.shape == ()
    assert l.requires_grad


def test_FMPE():
    estimator = FMPE(3, 5)

    # Non-batched
    theta, x, t = randn(3), randn(5), torch.tensor(0.5)
    score = estimator(theta, x, t)

    assert score.shape == (3,)
    assert score.requires_grad

    # Batched
    theta, x, t = randn(256, 3), randn(256, 5), randn(256)
    score = estimator(theta, x, t)

    assert score.shape == (256, 3)

    # Mixed
    theta, x, t = randn(256, 3), randn(5), randn(())
    score = estimator(theta, x, t)

    assert score.shape == (256, 3)

    # Sample
    x = randn(32, 5)
    theta = estimator.flow(x).sample((8,))

    assert theta.shape == (8, 32, 3)

    # Log-density
    log_p = estimator.flow(x).log_prob(theta)

    assert log_p.shape == (8, 32)


def test_FMPELoss():
    estimator = FMPE(3, 5)
    loss = FMPELoss(estimator)

    theta, x = randn(256, 3), randn(256, 5)

    l = loss(theta, x)

    assert l.shape == ()
    assert l.requires_grad


def test_MetropolisHastings():
    log_f = lambda x: -(x**2).sum(dim=-1) / 2
    f = lambda x: torch.exp(log_f(x))

    # f
    sampler = MetropolisHastings(randn(32, 3), f=f, sigma=0.5)

    it = iter(sampler)
    x1 = next(it)
    x2 = next(it)

    assert x1.shape == x2.shape == (32, 3)
    assert not (x1 == x2).all()

    # log_f
    sampler = MetropolisHastings(randn(32, 3), log_f=log_f, sigma=0.5)

    xs = list(sampler(256, burn=128, step=8))

    assert len(xs) == (256 - 128) // 8
    assert not (xs[0] == xs[1]).all()



r"""General purpose helpers."""

import numpy as np
import torch
import torch.nn as nn

from torch import Tensor
from torch.optim import Optimizer
from typing import *


class GDStep(object):
    r"""Creates a callable that performs gradient descent (GD) optimization steps
    for parameters :math:`\phi` with respect to differentiable loss values.

    The callable takes a scalar loss :math:`l` as input, performs a step

    .. math:: \phi \gets \text{GD}(\phi, \nabla_{\!\phi} \, l)

    and returns the loss, detached from the computational graph. To prevent invalid
    parameters, steps are skipped if not-a-number (NaN) or infinite values are found
    in the gradient. This feature requires CPU-GPU synchronization, which could be a
    bottleneck for some applications.

    Arguments:
        optimizer: An optimizer instance (e.g. :class:`torch.optim.SGD`).
        clip: The norm at which the gradients are clipped. If :py:`None`,
            gradients are not clipped.
    """

    def __init__(self, optimizer: Optimizer, clip: float = None):

        self.optimizer = optimizer
        self.parameters = [
            p
            for group in optimizer.param_groups
            for p in group['params']
        ]
        self.clip = clip

    def __call__(self, loss: Tensor) -> Tensor:
        if loss.isfinite().all():
            self.optimizer.zero_grad()
            loss.backward()

            if self.clip is None:
                self.optimizer.step()
            else:
                norm = nn.utils.clip_grad_norm_(self.parameters, self.clip)
                if norm.isfinite():
                    self.optimizer.step()

        return loss.detach()


def gridapply(
    f: Callable[[Tensor], Tensor],
    domain: Tuple[Tensor, Tensor],
    bins: Union[int, Sequence[int]] = 128,
    batch_size: int = 4096,
) -> Tuple[Tensor, Tensor]:
    r"""Evaluates a function :math:`f(x)` over a multi-dimensional domain split
    into grid cells. Instead of evaluating the function cell by cell, batches are
    given to the function.

    Arguments:
        f: A function :math:`f(x)`.
        domain: A pair of lower and upper domain bounds.
        bins: The number(s) of bins per dimension.
        batch_size: The size of the batches given to the function.

    Returns:
        The domain grid and the corresponding values.

    Example:
        >>> f = lambda x: -(x**2).sum(dim=-1) / 2
        >>> lower, upper = torch.zeros(3), torch.ones(3)
        >>> x, y = gridapply(f, (lower, upper), bins=8)
        >>> x.shape
        torch.Size([8, 8, 8, 3])
        >>> y.shape
        torch.Size([8, 8, 8])
    """

    lower, upper = domain

    # Shape
    dims = len(lower)

    if type(bins) is int:
        bins = [bins] * dims

    # Create grid
    coordinates = []

    for l, u, b in zip(lower, upper, bins):
        step = (u - l) / b
        ticks = torch.linspace(l, u - step, b).to(step) + step / 2
        coordinates.append(ticks)

    grid = torch.cartesian_prod(*coordinates).reshape(-1, dims)

    # Evaluate f(x) on grid
    y = [f(x) for x in grid.split(batch_size)]
    y = torch.cat(y)

    return grid.reshape(*bins, -1), y.reshape(*bins, *y.shape[1:])
